# A set of useful functions to help navigate the murky waters of radio 
# interferometry packages
import os
import sys
from astropy.io import fits as pyfits
import numpy
import math
import pyrap.tables
import subprocess
import im
import tempfile
import ms
import std
import glob
import time
import Owlcat.FitsTool as fitstool

# Load some Pyxis functionality
from Pyxis.ModSupport import *

# register ourselves with Pyxis and define the superglobals
register_pyxis_module(superglobals="MS LSM")
register_pyxis_module()

rm_fr = x.rm.args("-fr")

def fits2casa (input,output):
    """Converts FITS image to CASA image."""
    if exists(output):
        rm_fr(output)
    if not exists(input):
        abort("$input does not exist")
    std.runcasapy("importfits('$input','$output',overwrite=True)")
#    x.imagecalc("in='$input'","out=$output",split_args=False)

def make_threshold_mask (input="${im.RESTORED_IMAGE}",threshold=0,output="$im.MASK_IMAGE",high=1,low=0):
    """
    Makes a mask image by thresholding the input image at a given value. The output image is a copy of the input image,
    with pixel values of 'high' (1 by default) where input pixels are >= threshold, and 'low' (0 default) where pixels are <threshold.
    """
    input,output = interpolate_locals("input output")
    ff = pyfits.open(input)
    d = ff[0].data
    d[d<threshold] = low
    d[d>threshold] = high
    ff.writeto(output,clobber=True)
    info("made mask image $output by thresholding $input at %g"%threshold)

document_globals(make_threshold_mask,"im.RESTORED_IMAGE im.MASK_IMAGE");

define("COPY_IMAGE_TO_Template", "${MS:BASE}.imagecopy.fits","container for image copy")

def make_empty_image (msname="$MS",image="${COPY_IMAGE_TO}",channelize=None,**kw0):
    msname,image = interpolate_locals("msname image")
    # setup imager options
    kw0.update(dict(ms=msname,channelize=channelize,dirty=True,dirty_image=image,restore=False,
                   select="ANTENNA1==0 && ANTENNA2==1"))
    import im.lwimager
    im.lwimager.make_image(**kw0);
    # sometime this behaves funny, so leave nothing to chance
    hdu = pyfits.open(image)
    hdu[0].data[...] = 0
    hdu.writeto(image,clobber=True)
    info("created empty image $image")


# Borrow some fuctions from Owlcat/FitsTool

combine_fits = fitstool.stack_planes

splitfits = fitstool.unstack_planes

reorder_fits_axes = fitstool.reorder
    

def addcol(msname='$MS',colname=None,shape=None,
           data_desc_type='array',valuetype=None,init_with=0,**kw):
    """ add column to MS 
        msanme : MS to add colmn to
        colname : column name
        shape : shape
        valuetype : data type 
        data_desc_type : 'scalar' for scalar elements and array for 'array' elements
        init_with : value to initialise the column with 
    """
    msname, colname = interpolate_locals('msname colname')
    tab = pyrap.tables.table(msname,readonly=False)

    try: 
        tab.getcol(colname)
        info('Column already exists')

    except RuntimeError:
        info('Attempting to add %s column to %s'%(colname,msname))
        from pyrap.tables import maketabdesc
        valuetype = valuetype or 'complex'

        if shape is None: 
            dshape = list(tab.getcol('DATA').shape)
            shape = dshape[1:]

        if data_desc_type=='array':
            from pyrap.tables import makearrcoldesc
            coldmi = tab.getdminfo('DATA') # God forbid this (or the TIME) column doesn't exist
            coldmi['NAME'] = colname.lower()
            tab.addcols(maketabdesc(makearrcoldesc(colname,init_with,shape=shape,valuetype=valuetype)),coldmi)

        elif data_desc_type=='scalar':
            from pyrap.tables import makescacoldesc
            coldmi = tab.getdminfo('TIME')
            coldmi['NAME'] = colname.lower()
            tab.addcols(maketabdesc(makescacoldesc(colname,init_with,valuetype=valuetype)),coldmi)

        info('Column added successfuly.')

        if init_with:
            nrows = dshape[0]

            rowchunk = nrows//10 if nrows > 1000 else nrows
            for row0 in range(0,nrows,rowchunk):
                nr = min(rowchunk,nrows-row0)
                dshape[0] = nr
                tab.putcol(colname,numpy.ones(dshape,dtype=valuetype)*init_with,row0,nr)

    tab.close()


def toJy(val):
    _convert = dict(m=1e-3,u=1e-6,n=1e-9)
    unit = val.lower().split('jy')[0][-1]
    if str.isalpha(unit) :
        val= val.lower().split(unit+'jy')[0]
        return float(val)*_convert[unit]
    else:
        return float(val.lower().split('jy')[0])


def toDeg(val):
    """Convert angle to Deg. returns a float. val must be in form: 2arcsec, 2arcmin, or 2rad"""
    import math
    _convert = {'arcsec':3600.,'arcmin':60.,'rad':math.pi/180.,'deg':1.}
    val = val or cellsize
    ind = 1
    if not isinstance(val,str): 
        raise ValueError('Angle must be a string, e.g 10arcmin')
    for i,char in enumerate(val):
            if char is not '.':
                 try: 
                      int(char)
                 except ValueError:
                     ind = i
                     break
    a,b = val[:ind],val[ind:]
    try: 
        return float(a)/_convert[b]
    except KeyError: 
        abort('Could not recognise unit [%s]. Please use either arcsec, arcmin, deg or rad'%b)


def findImager(path,imager_name=None):
    """ Find imager"""
    ispath = len(path.split('/'))>1
    if ispath :
        if os.path.exists(path): 
            return path
        else : 
            return False
    # Look in system path
    check_path = subprocess.Popen(['which',path],stderr=subprocess.PIPE,stdout=subprocess.PIPE)
    stdout = check_path.stdout.read().strip()
    if stdout : 
        return path
#   # Check aliases
#   ##TODO: [@sphe] Potential issues with looking in aliases. Might need to review this
#   stdout = subprocess.Popen(['grep',path,'$HOME/.bash_aliases'],stderr=subprocess.PIPE,stdout=subprocess.PIPE)
#   err,out = stdout.stderr.read(),stdout.stdout.read()
#   if err:
#       return False
#   elif out: 
#       path = out.split('=')[-1].strip("'")
#       path = path.split('#')[0] # Incase there is a comment im the line
#       return path
    else:
        return False # Exhausted all sensible options, give up. 


    

def gen_run_cmd(path,options,suf='',assign='=',lv_str=False,pos_args=None):
    """ Generate command line run command """

    pos_args = pos_args or []
    run_cmd = '%s '%path

    for key,val in options.items():
        if val is not None:
            if isinstance(val,str) and lv_str:
                val = '"%s"'%val
            if isinstance(val,bool) and suf:
                if val:
                    run_cmd+='%s%s '%(suf,key)
            else:
                run_cmd+='%s%s%s%s '%(suf,key,assign,val)

    for arg in pos_args:
        run_cmd += '%s '%arg
    return run_cmd


def icasa(taskname, mult=None, loadthese=[], **kw0):
    """ 
      runs a CASA task given a list of options.
      A given task can be run multiple times with a different options, 
      in this case the options must be parsed as a list/tuple of dictionaries via mult, e.g 
      icasa('exportfits',mult=[{'imagename':'img1.image','fitsimage':'image1.fits},{'imagename':'img2.image','fitsimage':'image2.fits}]). 
      Options you want be common between the multiple commands should be specified as key word args.
    """

    # create temp directory from which to run casapy
    td = tempfile.mkdtemp(dir='.')
    # we want get back to the working directory once casapy is launched
    cdir = os.path.realpath('.')
    
    casapy = xro.which("casapy").strip() or xro.which("casa").strip() 
    # load modules in loadthese
    _load = ""
    if "os" not in loadthese or "import os" not in loadthese:
        loadthese.append("os")

    if loadthese:
        exclude = [line for line in loadthese if line.startswith("import") or line.startswith("from")]
        for line in loadthese:
            if line not in exclude:
                line = "import %s"%line
            _load += "%s\n"%line

    if mult:
        if isinstance(mult,(tuple,list)):
            for opts in mult:
                opts.update(kw0)
        else:
            mult.upadte(kw0)
            mult = [mult]
    else:
        mult = [kw0]

    run_cmd = """ """
    for kw in mult:
        task_cmds = ''
        for key,val in kw.items():
            if isinstance(val,str):
                 val = '"%s"'%val
            task_cmds += '\n%s=%s'%(key,val)
        run_cmd += """
%s

os.chdir('%s')
taskname = '%s'
%s
go()

"""%(_load,cdir,taskname,task_cmds)

    tf = tempfile.NamedTemporaryFile(suffix='.py')
    tf.write(run_cmd)
    tf.flush()
    t0 = time.time()
    # all logging information will be in the pyxis log files 
    x.sh('cd $td && $casapy --nologger --log2term --nologfile -c %s'%(tf.name))

    # log taskname.last 
    task_last = '%s.last'%taskname
    if exists(task_last):
        with open(task_last,'r') as last:
            info('${taskname}.last is: \n %s'%last.read() )

    # remove temp directory. This also gets rid of the casa log files; so long suckers!
    rm_fr(td,task_last)
    tf.close()



from Pyxis.ModSupport import *

register_pyxis_module();

import socket
import glob
import time
import os.path

define('PROJECT',"meerkat-7-gazing","default GCE project name")
define('ZONE',"europe-west1-b","default GCE compute zone")

# gcloud executable
# includes PROJECT and ZONE in the command line
gc = x.gcloud.args(before="compute --project $PROJECT",after="--zone $ZONE");
gc1 = x.gcloud.args(before="compute --project $PROJECT");
gco = xo.gcloud.args(before="compute --project $PROJECT",after="--zone $ZONE");
gcr = xr.gcloud.args(before="compute --project $PROJECT",after="--zone $ZONE");
gcr1 = xr.gcloud.args(before="compute --project $PROJECT");
gcro = xro.gcloud.args(before="compute --project $PROJECT",after="--zone $ZONE");

# gsutil cp executables
gcp = x.gsutil.args("cp");
gcpo = xo.gsutil.args("cp");

define('VM_SNAPSHOT',"oms-papino-*","snapshot on which VM boot disks are to be based")
define('VM_TYPE',"n1-standard-1","default VM type")

define('VM_USER',E.USER,"default username to be used on remote machine")

define('VM_NUM',1,"VM serial number, included in instance name")
define('VM_NAME_Template',"${VM_USER}-"+socket.gethostname().replace(".","-").lower()+"-$VM_NUM","default VM instance name");
define('VM_OUTPUT_BUCKET_Template',"gs://$VM_USER/outputs/$VM_NAME-$OUTDIR","default cloud storage destination for output products"); 

define('VM_DATADISKSIZE',200,"default data disk size (in Gb) for attach_disk()")

define('VM_PROVISION_SCRIPT','./pyxis-provision.sh','provisioning script to run on remote VM')

def provision_vm (vmname="$VM_NAME"):
  """Waits for specified VM to come up and runs provisioning script""";
  name = interpolate_locals("vmname");
  for attempt in range(1,11):
    if gco("ssh $name --command 'if [ -x $VM_PROVISION_SCRIPT ]; then $VM_PROVISION_SCRIPT; fi'") is 0:
      break;
    if attempt is 1 and name not in get_vms():
      abort("no such VM $name")
    warn("VM $name is not up yet (attempt #$attempt), waiting for 5 seconds to retry");
    time.sleep(5);
  else:
    abort("failed to connect to VM $name after $attempt tries")
  info("VM $name has been provisioned ")
  return True;

document_globals(provision_vm,"VM_NAME VM_PROVISION_SCRIPT PROJECT ZONE");
  
def _version_suffix (x,sep='-'):
  """Helper function: given a string such as "foo-N", returns N as integer, or 0 if string does not match.""";
  try:
    return int(x.rsplit(sep,1)[-1]);
  except:
    return 0;

## create VM
def init_vm (vmname="$VM_NAME",vmtype="$VM_TYPE",
             autodelete=True,reuse_boot=True,autodelete_boot=None,propagate=True,**kw):
  """Creates a GCE VM instance.
  vmname:           instance name
  vmtype:           VM type (use gce.list_machine_types() to get a list)
  autodelete:       if True, deletes any existing instance with same name
  reuse_boot:       if True and a boot disk with same name exists, reuses that
  autodelete_boot:  if True, boot disk will be auto-deleted when VM is shut down
  propagate:        if True, calls propagate_scripts() to copy Pyxis scripts from current
                    directory to VM. Can be set to a directory name to propagate to a 
                    specific directory on the remote machine.
  attach_XXX:       create additional disks. Use attach_XXX=N to create a disk of N Gb and 
                    attach to VM under /XXX (equivalent to calling attach_disk('XXX',size=N)). 
                    attach_XXX=dict(...) will call attach_disk('XXX',...)
""";
  name,vmtype = interpolate_locals("vmname vmtype");
  # check if VM exists and needs to be deleted
  if autodelete and name in get_vms():
    warn("deleting existing VM $name");
    delete_vm(name,disks=False);
  # check if a boot disk needs to be created
  disks = get_disks();
  if name in disks:
    if reuse_boot:
      info("boot disk $name already exists, reusing (disable with reuse_boot=False)")
      if autodelete_boot is None:
        info("boot disk $name will not be auto-deleted when VM is destroyed")
        autodelete_boot = False;
    else:
      gc("disks delete $name --quiet")
      del disks[name];
  if name not in disks:
    snapshot = VM_SNAPSHOT;
    if '*' in snapshot:
      matching = sorted(list(get_snapshots(snapshot).keys()),
        lambda a,b:-cmp(_version_suffix(a),_version_suffix(b)));
      snapshot = matching[0];
      info("using latest snapshot $snapshot");
    gc("disks create $name --source-snapshot $snapshot")
    if autodelete_boot is None:
      info("boot disk $name will be auto-deleted when VM is destroyed")
      autodelete_boot = True;
  # create VM
  scopes = "--scopes storage-rw"
  gc("instances create $name --machine-type $vmtype --disk name=$name mode=rw boot=yes auto-delete=%s $scopes"%("yes" if autodelete_boot else "no"));
  info("created VM instance $name, type $vmtype")
  # run provisioning script
  provision_vm(name);
  # attach disks
  for key,value in kw.items():
    if key.startswith("attach_"):
      if isinstance(value,dict):
        attach_disk(key[len("attach_"):],vmname=vmname,**value);
      elif isinstance(value,int):
        attach_disk(key[len("attach_"):],size=value,vmname=vmname);
      else:
      	raise TypeError("unknown data type for %s"%key);
  # provision with pyxis scripts in specified directory
  if propagate:
    propagate_scripts(name,dir=propagate if isinstance(propagate,str) else "");

document_globals(init_vm,"VM_NAME VM_TYPE VM_SNAPSHOT VM_DATADISKSIZE PROJECT ZONE");

def rsh (command,vmname='$VM_NAME',bg=False):
  """Executes a command on a VM""";
  command,vmname = interpolate_locals("command vmname");
  if bg:
    gc("ssh $vmname --command 'screen -L -md $command'")
  else:
    gc("ssh $vmname --command '$command'")

document_globals(rsh,"VM_NAME PROJECT ZONE");

def rpyxis (command,vmname='$VM_NAME',dir=None,bg=False,wrapup=False):
  """Executes a pyxis command on a VM
  command:    command(s) to execute
  vmname:     instance name
  dir:        change into directory before running command. Default will 
              run on home directory on VM.
  bg:         if False, waits until command has finished. If True, launches
              command remotely (under screen) and returns
  wrapup:     if True, remote will call gce.wrapup() after command has finished.
              This is generally used to copy output to GCE storage and kill the VM.
  """;
  command,vmname = interpolate_locals("command vmname");
  cd = II("cd $dir;") if dir else "";
  wrap = II("--wrapup gce.wrapup") if wrapup else "";
  if bg:
    gc("ssh $vmname --command 'screen -L -md bash -i -c \"$cd pyxis gce.VM_NAME=$vmname $command $wrap\"'")
  else:
    gc("ssh $vmname --command 'bash -i -c \"$cd pyxis gce.VM_NAME=$vmname $command $wrap\"'")

document_globals(rpyxis,"VM_NAME PROJECT ZONE");

def propagate_scripts (vmname="$VM_NAME",dir=""):
  """Copies pyxis-*.{py,conf} from current directory to VM.
  If 'dir' is specified, copies to specific directory, else copies to home directory.
  """
  name,dir = interpolate_locals("vmname dir");
  # make sure the machine is ready -- retry the file copy until we succeed
  files = " ".join(list(glob.glob("pyxis-*py")) + list(glob.glob("pyxis-*.conf")));
  if files:
    gc("copy-files $files $name:$dir");
    info("propagated $files to VM");
  else:
    info("no pyxis-*{py,conf} files in current directory, nothing to propagate")

document_globals(propagate_scripts,"VM_NAME PROJECT ZONE");

def _remote_attach_disk (diskname,mount,clear):
  """Internal helper function run by remote VM to attach a disk""";
  if not os.path.exists(mount):
    x.sh("sudo mkdir $mount");
  x.sh("sudo /usr/share/google/safe_format_and_mount -m 'mkfs.ext4 -F' /dev/disk/by-id/google-$diskname $mount");
  x.sh("sudo chown $VM_USER.$VM_USER $mount");
  if clear:
    x.sh("sudo rm -fr $mount/*");
  # make symlink, if mounting at root level 
  mm = os.path.realpath(mount).split("/");
  if len(mm) < 3 and not os.path.exists(mm[-1]):
    gc("ln -s $mount");

def attach_disk (mount="data",diskname="${vmname}-$mount",vmname="$VM_NAME",
                 size=None,snapshot=None,ssd=False,
                 init=False,clear=False,mode="rw",autodelete=False):
  """Attached an extra disk to a VM (creating the disk as needed).
  mount:        where to mount the disk, e.g. mount='data' mounts under /data
  diskname:     provide explicit name for the disk (otherwise auto-named as "vmname-mount"). 
                Can be used to attach previously created disks.
  vmname:       name of VM instance
  size:         disk size for new disks, VM_DATADISKSIZE by default
  snapshot:     init disk from a named snapshot (else create empty disk)
  ssd:          if True, creates (faster but more expensive) SSD disk
  init:         if True and disk already exists, deletes and re-creates it. If False, reuses.
  clear:        if True and disk already exists, wipes its data.
  mode:         default "rw". Use "ro" to attach as read-only.
  autodelete:   if True, disk will be deleted when the VM is shut down
  """
  name,diskname,disksize,mount = interpolate_locals("vmname diskname size mount")
  diskname = diskname.lower().replace("/","");
  disks = get_disks();
  if diskname in disks and init:
    info("disk $diskname exists and init=True, recreating")
    gc("disks delete $diskname");
    del disks[diskname];
  if diskname not in disks:
    disktype = "pd-ssd" if ssd else "pd-standard";
    info("disk $diskname does not exist, creating type $disktype${ size <disksize> Gb}${ from snapshot <snapshot}")
    if not snapshot:
      size = VM_DATADISKSIZE;
    gc("disks create $diskname ${--size <disksize} --type $disktype ${--source-snapshot <snapshot}");
    clear = False;
  # attach disk to VM
  gc("instances attach-disk $name --disk $diskname --mode $mode --device-name $diskname")
  if autodelete:
    gc("instances set-disk-auto-delete --auto-delete $name --disk $diskname")
  # execute rest on remote
  rpyxis('gce._remote_attach_disk[:$diskname:,:$mount:,$clear]',name);
  info("attached disk $diskname as $name:$mount ($mode)")

document_globals(attach_disk,"VM_NAME VM_DATADISKSIZE PROJECT ZONE")

def detach_disk (mount="data",vmname="$VM_NAME",diskname="${vmname}-$mount"):
  """Detaches disk from VM instance
  mount:      mountpoint of disk.
  vmname:     name of VM instance.
  diskname:   explicit disk name, formed as vmname-mount by default.
  """;
  name,diskname,mount = interpolate_locals("vmname diskname mount")
  gc("instances detach-disk $name --disk $diskname")
  info("detached disk $diskname from VM $name")

document_globals(detach_disk,"VM_NAME PROJECT ZONE")

def get_vms ():
  """returns list of all VMs in zone""";
  return dict([ x.split(None,1) for x in gcr("instances list").split("\n")[1:] if x])
document_globals(get_vms,"PROJECT ZONE");

def list_vms ():
  """prints list of all VMs in zone""";
  gc("instances list");
document_globals(list_vms,"PROJECT ZONE");

def get_snapshots (pattern="*"):
  """returns list of all snapshots in zone""";
  a = [ x.split(None,1) for x in gcr1("snapshots list").split("\n")[1:] if x ];
  return dict([ x for x in a if fnmatch.fnmatch(x[0],pattern) ]);
document_globals(get_snapshots,"PROJECT ZONE");

def list_snapshots ():
  """prints list of all snapshots in zone""";
  gc1("snapshots list");
document_globals(list_snapshots,"PROJECT ZONE");

def get_disks ():
  """returns list of all disks in zone""";
  return dict([ x.split(None,1) for x in gcr("disks list").split("\n")[1:] if x])
document_globals(get_disks,"PROJECT ZONE");

def list_disks ():
  """prints list of all disks in zone""";
  gc("disks list");
  # vms = get_disks().items();
  # for name,data in sorted(vms):
  #   info("disk $name: $data");
document_globals(list_disks,"PROJECT ZONE");

def list_machine_types ():
  """prints list of all VM types available in zone (i.e. possible VM_TYPE settings)""";
  gc("machine-types list");
document_globals(list_machine_types,"PROJECT ZONE");

def delete_disk (*disknames):
  """deletes the specified disk(s)"""
  for disk in disknames:
    gc("disks delete $disk --quiet");
document_globals(delete_disk,"PROJECT ZONE");

def delete_vm (vmname="$VM_NAME",disks=True):
  """Deletes a GCE VM instance. If disks=True, deletes associated data disks.""";
  name = interpolate_locals("vmname");
  gco("instances delete $name --quiet");
  info("deleted VM instance $name" + ("; deleting asociated disks" if disks else ""));
  if disks:
    for key,value in get_disks().items():
      if key.startswith(name+"-"):
        gc("disks delete $key --quiet");
document_globals(delete_vm,"VM_NAME PROJECT ZONE");


def wrapup ():
  """Helper function run on remote VMs to wrap up (i.e. if wrapup=True is given to gce.rpyxis()). Copies
  outputs to cloud storage and shuts down the VM.
  """
  files = [ f for f in glob.glob("/var/log/syslog*") + 
            glob.glob(II("$OUTDIR/*txt")) +
            glob.glob(os.path.expanduser("~/screenlog.*")) if exists(f) ];
  if files:
    gcpo("%s $VM_OUTPUT_BUCKET"%" ".join(files));
  x.sh("sudo poweroff")
document_globals(wrapup,"VM_NAME VM_OUTPUT_BUCKET OUTDIR");



#!/usr/bin/env python

import os
from distutils.core import setup

install_requires = [
      'astropy',
      # 'timba', not available on pypi
      'matplotlib',
      'python_casacore',
      'numpy',
      'future',
      'six',
]

setup(name='astro-pyxis',
      version='1.6.1',
      description='Python Extensions for astronomical Interferometry Scripting',
      author='Oleg Smirnov',
      author_email='Oleg Smirnov <osmirnov@gmail.com>',
      url='https://github.com/ska-sa/pyxis',
      packages=['Pyxis', 'Pyxides', 'Pyxides._utils', 'Pyxides.im'],
      install_requires=install_requires,
      scripts=['Pyxis/bin/' + i for i in os.listdir('Pyxis/bin')],
     )



from Pyxis.ModSupport import *

import imager,std

from astropy.io import fits as pyfits
import Tigger
import im.argo

register_pyxis_module(superglobals="OUTFILE");

tigger_restore  = x("tigger-restore")
tigger_convert  = x("tigger-convert")
tigger_tag      = x("tigger-tag")

v.define("LSM","lsm.lsm.html",
  """current local sky model""");

define("LSM_TDL_Template","tiggerlsm.filename=$LSM",
  """TDL option for selecting current lsm""");
define("LSMREF","",
  """reference LSM (for transferring tags, etc.)""");
  
define('PYBDSM_OUTPUT_Template',"${OUTFILE}_pybdsm.lsm.html","""output LSM file for pybdsm search""");    
define('PYBDSM_POLARIZED',None,"""set to True to run pybdsm in polarized mode""");
define('PYBDSM_OPTIONS',{},"Extra options given to pybdsm");
_pybdsm = x.pybdsm;

define('CLUSTER_DIST',0,
  """source clustering distance, arcsec. If 0, then CLUSTER_DIST_BEAMS is used instead.""");
define('CLUSTER_DIST_BEAMS',3,
  """source clustering distance, in units of PSF size (measured as (BMAJ+BMIN)/2). If BMAJ/BMIN is not defined,
  this falls back to 60 arcsec.""");
define('MIN_EXTENT',0,
  """minimum Gaussian source extent, arcsec; sources smaller than this will be converted to point sources""");

def pybdsm_search (image="${imager.RESTORED_IMAGE}",output="$PYBDSM_OUTPUT",pol='$PYBDSM_POLARIZED',
  select=None,center=None,
  threshold=None,pbexp=None,**kw):
  """Runs pybdsm on the specified 'image', converts the results into a Tigger model and writes it to 'output'.
  Use 'threshold' to specify a non-default threshold (thresh_isl and thresh_pix).
  Use 'pol' to force non-default polarized mode.
  Use 'pbexp' to supply a primary beam expression (passed to tigger-convert), in which case the output model will contain
  intrinsic fluxes.
  Use 'select' to apply a selection string on the new model (e.g. "I.gt.0.001")
  Use 'center' to set the centre of the model (useful for e.g. selecting on radius), use a string e.g. Xdeg,Ydeg
  """
  image,output,pol,center = interpolate_locals("image output pol center");
  makedir(v.DESTDIR);
  # setup parameters
  gaul = II("${output:BASEPATH}.gaul");
  # info("PyBDSM filenames are $output $gaul");
  # start with default PYBDSM options
  opts = PYBDSM_OPTIONS.copy();
  opts.update(kw);
  # override with explicit arguments
  if threshold:
    opts['thresh_pix'] = threshold;
  if pol is not None:
    opts['polarisation_do'] = is_true(pol);
  pol = opts.get('polarisation_do',False);
  opts['quiet'] = True;
  # run pybdsm
  info("running PyBDSM process_image($image,%s)"%",".join(sorted([ "%s=%s"%x for x in opts.items() ])));
  from lofar import bdsm
  img = bdsm.process_image(image,**opts);
  info("writing PyBDSM gaul catalog");
  img.write_catalog(outfile=gaul,format='ascii',catalog_type='gaul',clobber=True);
  # add log to output
  logfile = II("${output:BASEPATH}.pybdsm.log");
  if exists(logfile):
    info("PyBDSM log output follows:");
    for line in file(logfile):
      print("     ",line);
  else:
    warn("PyBDSM log $logfile not found");
  # set clustering parameter from beam size
  cluster = CLUSTER_DIST;
  if not cluster:
    hdr = pyfits.open(image)[0].header;
    # BMAJ/BMIN is in degrees -- convert to seconds, or fall back to 60" if not set
    cluster = 1800*(hdr.get('BMAJ',0)+hdr.get('BMIN',0))*CLUSTER_DIST_BEAMS or 60;
  # convert catalog
  if pbexp:
    args = [ "--primary-beam",pbexp,"--app-to-int" ]
  else:
    args = []
  if select:
    args += [ "--select",select ];
  if center:
    args += [ "--center",center ]
  
  ## leaving this for now but it seems overly complicated. Better to add an option
  ## to tigger-convert to eliminate sources with NaN positions
  verifyGaulModel(gaul)
  
  tigger_convert(gaul,output,"-t","Gaul",
    "-f","--rename",
    "--cluster-dist",cluster,
    "--min-extent",MIN_EXTENT,
    split_args=False,
    *args);
    
document_globals(pybdsm_search,"PYBDSM_* imager.RESTORED_IMAGE CLUSTER_* MIN_EXTENT");

def verifyGaulModel(gaullsm):
  """Check all sources in a gaul file are in valid locations before running tigger
  convert. Useful when images are 'all-sky' and have undefined regions.
  """
  falseSources=0
  olsm=''
  fh=open(gaullsm,'r')
  for ll in fh.readlines():
    cll=' '.join(ll.split())
    if cll=='' or cll.startswith('#'):
      olsm+=ll
      continue
    lineArray=cll.split(' ')
    if math.isnan(float(lineArray[4])): falseSources+=1
    else: olsm+=ll
  fh.close()

  fh=open(gaullsm,'w')
  fh.write(olsm)
  fh.close()


def transfer_tags (fromlsm="$LSMREF",lsm="$LSM",output="$LSM",tags="dE",tolerance=60*ARCSEC):
  """Transfers tags from a reference LSM to the given LSM. That is, for every tag
  in the given list, finds all sources with those tags in 'fromlsm', then applies 
  these tags to all nearby sources in 'lsm' (within a radius of 'tolerance'). 
  Saves the result to an LSM file given by 'output'.
  """
  fromlsm,lsm,output,tags = interpolate_locals("fromlsm lsm output tags");
  # now, set dE tags on sources
  tagset = frozenset(tags.split());
  info("Transferring tags %s from %s to %s (%.2f\" tolerance)"%(",".join(tagset),fromlsm,lsm,tolerance/ARCSEC));
  import Tigger
  refmodel = Tigger.load(fromlsm);
  model = Tigger.load(lsm);
  # for each dE-tagged source in the reference model, find all nearby sources
  # in our LSM, and tag them
  for src0 in refmodel.getSourceSubset(",".join(["="+x for x in tagset])):
    for src in model.getSourcesNear(src0.pos.ra,src0.pos.dec,tolerance=tolerance):
      for tag in tagset:
        tagval = src0.getTag(tag,None);
        if tagval is not None:
          if src.getTag(tag,None) != tagval:
            src.setTag(tag,tagval);
            info("setting tag %s=%s on source %s (from reference source %s)"%(tag,tagval,src.name,src0.name))
  model.save(output);

  

CC_RESCALE = 1.  
CC_IMAGE_Template = "${OUTFILE}_ccmodel.fits"

def add_ccs (lsm="$LSM",filename="${imager.MODEL_IMAGE}",
             cc_image="$CC_IMAGE",srcname="ccmodel",output="$LSM",zeroneg=True,scale=None,pad=1):
  """Adds clean components from the specified FITS image 'filename' to the sky model given by 'lsm'.
  Saves the result to an LSM file given by 'output'.
  The CC image is copied to 'cc_image', optionally rescaled by 'scale', and optionally has negative pixels reset to zero (if zeroneg=True).
  'srcname' gives the name of the resulting LSM component.
  'pad' gives the padding attribute of the LSM component, use e.g. 2 if CC image has significant signal towards the edges.
  """;
  lsm,filename,cc_image,srcname,output = interpolate_locals("lsm filename cc_image srcname output");
  info("adding clean components from $filename ($cc_image), resulting in model $output");
  # rescale image
  ff = pyfits.open(filename);
  ff[0].data *= (scale if scale is not None else CC_RESCALE);
  if zeroneg:
    ff[0].data[ff[0].data<0] = 0;
  ff.writeto(cc_image,clobber=True);
  tigger_convert(lsm,output,"-f","--add-brick","$srcname:$cc_image:%f"%pad);

document_globals(add_ccs,"MODEL_CC_*");  


def pointify (lsm="$LSM",output="$LSM",name=""):
  """Replaces names sources with point sources""";
  lsm,output,name = interpolate_locals("lsm output name");
  model = Tigger.load(lsm);
  src = model.findSource(name);
  info("Setting source $name in model $lsm to point source, saving to $output");
  src.shape = None;
  model.save(output);

document_globals(add_ccs,"MODEL_CC_*");  


define('SOFIA_CFG','${OUTFILE}_sofia_conf.txt','Name of auto generated sofia config file')
define('SOFIA_PATH','sofia_pipeline.py','SoFia pipeline path (sofia_pipeline.py)')
define('SOFIA_OUTDIR','${v.DESTDIR}','SoFiA output directory')
define('SOFIA_OUT_PREFIX','${OUTFILE:BASE}sofia','Prefix for sofia outputs')
define('_SOFIA_DEFAULTS',{},'Dictionary of default SoFiA options')

_SOFIA_DEFAULTS = {'writeCat': {'writeASCII': 'true', 'parameters': "['*']", 'compress': 'false', 'basename': '', 'writeXML': 'false', 'writeSQL': 'false', 'outputDir': ''}, 'parameters': {'dilateThreshold': '0.02', 'optimiseMask': 'false', 'fitBusyFunction': 'false', 'dilatePixMax': '10', 'dilateChan': '1', 'dilateMask': 'false'}, 'smooth': {'kernelZ': '3.0', 'kernel': 'gaussian', 'kernelX': '3.0', 'kernelY': '3.0', 'edgeMode': 'constant'}, 'merge': {'radiusY': '3', 'radiusX': '3', 'radiusZ': '3', 'minSizeY': '3', 'minSizeX': '3', 'minSizeZ': '2'}, 'flag': {'regions': '[]'}, 'optical': {'storeMultiCat': 'false', 'sourceCatalogue': '', 'specSize': '1e+5', 'spatSize': '0.01'}, 'steps': {'doReliability': 'false', 'doWriteFilteredCube': 'false', 'doMom1': 'false', 'doParameterise': 'true', 'doFlag': 'false', 'doSmooth': 'false', 'doDebug': 'false', 'doCNHI': 'false', 'doWavelet': 'false', 'doCubelets': 'false', 'doWriteMask': 'false', 'doOptical': 'false', 'doSubcube': 'false', 'doMerge': 'true', 'doMom0': 'false', 'doScaleNoise': 'false', 'doSCfind': 'true', 'doWriteCat': 'true', 'doThreshold': 'false'}, 'wavelet': {'threshold': '5.0', 'scaleXY': '-1', 'positivity': 'false', 'iterations': '3', 'scaleZ': '-1'}, 'threshold': {'threshold': '4.0', 'rmsMode': 'std', 'clipMethod': 'relative', 'verbose': 'false'}, 'import': {'weightsFunction': '', 'subcubeMode': 'pixel', 'subcube': '[]', 'maskFile': '', 'inFile': '', 'weightsFile': ''}, 'CNHI': {'verbose': '1', 'qReq': '3.8', 'maxScale': '-1', 'medianTest': 'true', 'minScale': '5', 'pReq': '1e-5'}, 'reliability': {'threshold': '0.9', 'kernel': '[0.15,0.05,0.1]', 'parSpace': "['SNRsum','SNRmax','NRvox']", 'fMin': '0.0', 'makePlot': 'false'}, 'scaleNoise': {'scaleX': 'false', 'scaleY': 'false', 'scaleZ': 'true', 'statistic': 'mad', 'edgeX': '0', 'edgeY': '0', 'edgeZ': '0'}, 'SCfind': {'kernels': "[[ 0, 0, 0,'b'],[ 0, 0, 3,'b'],[ 0, 0, 7,'b'],[ 0, 0, 15,'b'],[ 3, 3, 0,'b'],[ 3, 3, 3,'b'],[ 3, 3, 7,'b'],[ 3, 3, 15,'b'],[ 6, 6, 0,'b'],[ 6, 6, 3,'b'],[ 6, 6, 7,'b'],[ 6, 6, 15,'b']]", 'maskScaleXY': '2.0', 'verbose': 'true', 'rmsMode': 'negative', 'edgeMode': 'constant', 'kernelUnit': 'pixel', 'threshold': '4.0', 'maskScaleZ': '2.0', 'sizeFilter': '0.0'}}

def sofia_search(fitsname='${im.RESTORED_IMAGE}',sofia_conf=None,
                 threshold=4,do_reliability=True,reliability=0.9,
                 merge=3,basename='$SOFIA_OUT_PREFIX',makeplot=True,
                 outdir='$SOFIA_OUTDIR',options={}):
    """ Runs SoFiA source finding pipeline. Only a few options are provided here. 
        For more eleborate settings, add options (as you would in a SoFiA configuarion file) 
        via the [options] dictionary or provide a SoFiA configuration file via [sofia_conf]
       -------
       fitsname : Name of fits map on which to run the source finder
       sofia_conf : SoFiA configuration file
       threshold : Peak threshold for source finder [in units of sigma above noise rms]
       do_reliability : Do reliability caltulations
       reliability : Reliability threshold. (0,1]
       merge : merge 
       options : extra options which will directly to sofia configuration file.
    """

    fitsname,basename,outdir = interpolate_locals('fitsname basename outdir')
    makedir(v.DESTDIR);

    # swap freq and stokes in fits hdr
    with pyfits.open(fitsname) as hdu:
        hdr = hdu[0].header
        ndim = hdr["NAXIS"]
        if ndim>3 and hdr["CTYPE4"].startswith("FREQ"):
            warn("Swapping STOKES and FREQ axes in IMAGE $fitsname. This is a SoFiA workaround")
            im.argo.reorder_fits_axes(fitsname, order=[1,2,4,3], outfile=fitsname)

    # use default SoFiA config if not specified
    if not sofia_conf:

        sofia_conf = II(SOFIA_CFG)
        _SOFIA_DEFAULTS['import']['inFile'] = fitsname
        _SOFIA_DEFAULTS['writeCat']['outputDir'] = outdir
        _SOFIA_DEFAULTS['writeCat']['basename'] = basename

        if os.path.exists(sofia_conf):
            x.sh(II('mv $sofia_conf ${sofia_conf}.old'))
        sofia_std = open(sofia_conf,'w')

        if threshold!=4 and isinstance(threshold,(int,float)):
            _SOFIA_DEFAULTS['threshold']['threshold'] = threshold
        if do_reliability and isinstance(do_reliability,bool):
            _SOFIA_DEFAULTS['steps']['doReliability'] = 'true'
            if isinstance(reliability,(int,float)):
                _SOFIA_DEFAULTS['reliability']['threshold'] = reliability
                _SOFIA_DEFAULTS['reliability']['makePlot'] = makeplot
        if merge!=3 and isinstance(merge,int):
            _SOFIA_DEFAULTS['merge']['mergeX'] = merge
            _SOFIA_DEFAULTS['merge']['mergeY'] = merge
            _SOFIA_DEFAULTS['merge']['mergeZ'] = merge

      # Update default SoFia configuration dictionary with user options
        for key,val in options.items():
            a,b = key.split('.')
            if a not in list(_SOFIA_DEFAULTS.keys()):
                abort('Option ${a}.${b} is not recognisable.')
            _SOFIA_DEFAULTS[a][b] = val

      # Generate sofia configuration file
        sofia_std.write('#Sofia autogen configuration file [pyxis]')
        for a,b in  _SOFIA_DEFAULTS.items():
            for key,val in b.items():
                sofia_std.write('\n%s.%s = %s'%(a,key,val))
        sofia_std.close()

    x.sh('$SOFIA_PATH $sofia_conf')

document_globals(sofia_search," im.RESTORED_IMAGE _SOFIA_DEFAULTS SOFIA_CFG SOFIA_OUTDIR SOFIA_PATH SOFIA_OUT_PREFIX")





fixuvw_casa = """msname = '$msname'
rowstep = ${rowstep?100000}
write_uvw = ${write_uvw?False}

# open main MS table, and ANTENNA and FIELD subtable
tb.open(msname+'/ANTENNA');
antpos = tb.getcol("POSITION").transpose()
# this will be a 2-vector for the direction of field 0 in MS

tb.open(msname+'/FIELD');
radec = tb.getcol("PHASE_DIR").transpose()

tb.open(msname,nomodify=not write_uvw)
# indices
ant1,ant2 = tb.getcol("ANTENNA1"),tb.getcol("ANTENNA2")
field = tb.getcol("FIELD_ID")
time = tb.getcol("TIME_CENTROID");
uvw0 = tb.getcol("UVW").transpose();
uvw = uvw0.copy();

import numpy,math

dq,dm = qa,me
nant = antpos.shape[0]

f0 = t0 = None

ant_uvw = antpos.copy();

# make list of ITRF baselines per antenna
ant_bl_itrf = [ dm.baseline('itrf',*[ dq.quantity(x,'m') for x in antpos[i,:]-antpos[0,:]]) for i in range(nant) ]

print "### Printing computed minus stored UVWs every %d rows. Hopefully the delta is small..."%rowstep;
for row,(t,f,a1,a2) in enumerate(zip(time,field,ant1,ant2)):
    if t != t0:
        epoch = dm.epoch("UTC",dq.quantity(t,"s"))
        dm.doframe(epoch)
    if f != f0:
        ra,dec = radec[f][0];
        dir0 = dm.direction('J2000',dq.quantity(ra,"rad"),dq.quantity(dec,"rad"));
        dm.doframe(dir0);
    # recompute all baselines for this timeslot
    if t != t0 or f != f0:
      # convert to J2000 UVWs
      ant_bl_j2000 = [ dm.touvw(x)[1]['value'] for x in ant_bl_itrf ];
      for i,bl in enumerate(ant_bl_j2000):
        ant_uvw[i,:] = bl; 
    t0,f0 = t,f
    uvw[row,:] = ant_uvw[a1,:] - ant_uvw[a2,:];
    if not row%rowstep:
      print "row %d delta-uvw "%row,uvw[row,:] - uvw0[row,:]

if write_uvw:
  print "Writing new UVW column";
  tb.putcol("UVW",uvw.transpose());
""";